type arithOp = Plus | Minus | Times | Divide | Modulo
type relOp = Eq | Neq | Lt | Lte | Gt | Gte
type logicOp = And | Or
type binOp = ArithOp of arithOp | RelOp of relOp | LogicOp of logicOp
type monOp = UnaryMinus | Negation
type const = BoolConst of bool | IntConst of int
type expr =
    Enclosure of expr
    | BinOp of expr * binOp * expr
    | MonOp of monOp * expr
    | Id of string
    | Const of const

let rec eval_expr =
    let eval_binOp =
        let eval_arithOp = function
            Plus -> "+"
            | Minus -> "-"
            | Times -> "*"
            | Divide -> "/"
            | Modulo -> "mod"
        in
        let eval_relOp = function
            Eq -> "="
            | Neq -> "!="
            | Lt -> "<"
            | Lte -> "<="
            | Gt -> ">"
            | Gte -> ">="
        in
        let eval_logicOp = function
            And -> "&&"
            | Or -> "||"
        in
        function
        ArithOp op -> eval_arithOp(op)
        | RelOp op -> eval_relOp(op)
        | LogicOp op -> eval_logicOp(op)
    in
    let eval_monOp = function
        UnaryMinus -> "-"
        | Negation -> "!"
    in
    let eval_const = function
        IntConst i -> string_of_int i
        | BoolConst b -> string_of_bool b
    in
    function
    Enclosure e -> "(" ^ eval_expr(e) ^ ")"
    | BinOp (e1, op, e2) -> eval_expr(e1) ^ " " ^ eval_binOp(op)
                            ^ " " ^ eval_expr(e2)
    | MonOp (op, e) -> eval_monOp(op) ^ eval_expr(e)
    | Id id -> id
    | Const c -> eval_const(c)
;;

(* (a) *)
print_endline (eval_expr (Enclosure (Id "a")));;
(* -a *)
print_endline (eval_expr (MonOp (UnaryMinus, Id "a")));;
(* a - b *)
print_endline (eval_expr (BinOp (Id "a", ArithOp Minus, Id "b")));;
(* a * (-b + 1) *)
let a_times b = BinOp (Id "a", ArithOp Times, b) in
let uminus a = MonOp (UnaryMinus, a) in
let plus a b = BinOp (a, ArithOp Plus, b) in
let one = Const (IntConst 1) in
print_endline (eval_expr (a_times (Enclosure (plus (uminus (Id "b")) one))));;
(* a = b && c *)
let b_and_c = BinOp (Id "b", LogicOp And, Id "c") in
print_endline (eval_expr (BinOp (Id "a", RelOp Eq, b_and_c)));;
